#ifndef __VERBOSE_TRACE_H__
#define __VERBOSE_TRACE_H__

#define GCC_VERBOSE_TRACE_ENV "GCC_VERBOSE_TRACE"
#define PR_HEAD fprintf (stderr, "--TRACE(%s)--> ", title)

bool gcc_enable_trace __attribute__((__weak__));

static inline void
trace_cat_file (const char *title, const char *fname)
{
  if (!gcc_enable_trace)
    return;

  PR_HEAD; fprintf (stderr, ">>>>>>>>>> begin '%s'\n", fname);

  FILE *f = fopen (fname, "r");
  if (!f)
    {
      PR_HEAD; fprintf (stderr, "cannot open '%s'\n", fname);
    }
  else
    {
      int c;
      while ((c = fgetc (f)) != EOF)
        fputc (c, stderr);

      fclose (f);
    }

  PR_HEAD; fprintf (stderr, "<<<<<<<<<< end of '%s'\n", fname);
}

/* Put the trace_begin() in the head of main(), and trace_end()
 * in the tail. So we can known the tool how being called.
 */
static inline bool
trace_begin (const char *title, int argc, char **argv)
{
  if (!getenv (GCC_VERBOSE_TRACE_ENV))
    return false;

  PR_HEAD; fprintf (stderr, "begin tracing.\n");
  gcc_enable_trace = true;

  /* Show the tool driver's common environments.
   */
#define PENV(name) { \
	  const char *env = getenv (name); \
	  if (env) { PR_HEAD; fprintf (stderr, "%s=%s\n", name, env); } \
	}
  PENV ("COLLECT_GCC");
  PENV ("OFFLOAD_TARGET_NAMES");
  PENV ("COLLECT_GCC_OPTIONS");
  PENV ("GCC_EXEC_PREFIX");
  PENV ("COMPILER_PATH");
  PENV ("LIBRARY_PATH");
  PENV ("LD_RUN_PATH");
  PENV ("PATH");

  /* Show main's argv[].
   */
  PR_HEAD; fprintf (stderr, "%s", argv[0]);
  char *at_file_name = 0;
  for (int i = 1; i < argc; ++i)
    {
      if (argv[i][0] == '@')
        at_file_name = &(argv[i][0]);
      fprintf (stderr, " %s", argv[i]);
    }
  fprintf (stderr, "\n");
  if (at_file_name && !at_file_name[0])
    trace_cat_file (title, at_file_name);

  fprintf (stderr, "\n");
  return true;
}

static inline void
trace_end (const char *title)
{
  if (!gcc_enable_trace)
    return;

  gcc_enable_trace = false;
  PR_HEAD; fprintf (stderr, "end of tracing.\n");
}

static inline void
trace_fork_execute (const char *title, const char *prog, char **argv)
{
  if (!gcc_enable_trace)
    return;

  PR_HEAD; fprintf (stderr, "%s", prog);
  int i = 1;
  while (argv[i])
    fprintf (stderr, " %s", argv[i++]);
  fprintf (stderr, "\n");
}

#ifdef __cplusplus
class verbose_trace {
  const char *_title; // TODO: use std::string instead.
  bool _enable;

public:
  verbose_trace (const char *title, int argc, char **argv)
  {
    _title = title;
    _enable = trace_begin (_title, argc, argv);
  }

  ~verbose_trace () { trace_end (_title); }

  bool enable() { return _enable; }
};
#endif /* __cplusplus */

#endif /* __VERBOSE_TRACE_H__ */
